from .base_model import BaseModel
from .storage_config_data import StorageConfigData
from .network_config_data import NetworkConfigData


class MachineData(BaseModel):
    hostname: str
    storage: StorageConfigData
    network: NetworkConfigData



from pathlib import Path

from pyinfra.api import deploy
from pyinfra.operations import (
    apt,
    files,
    server,
    systemd,
)

from aircraft.deploys.ubuntu.models import v1beta3
from aircraft.validators import validate_schema_version

deploy_dir = Path(__file__).parent


@deploy('Configure apache2')
def configure(state=None, host=None):
    supported_schema_versions = [
        v1beta3.HttpData,
    ]

    validate_schema_version(host.data.http, supported_schema_versions)

    apt.packages(
        name='Install package',
        packages=['apache2'],
        sudo=True,

        state=state, host=host,
    )

    files.directory(
        name=f'Ensure HTTP root dir {host.data.http.root_dir}',
        path=str(host.data.http.root_dir),
        present=True,
        recursive=True,
        sudo=True,

        state=state, host=host,
    )

    apache_conf = files.template(
        name='Render config file',
        src=str(deploy_dir / 'templates' / 'apache2-directory.conf.j2'),
        dest=str(Path('/etc') / 'apache2' / 'conf-available' / 'root.conf'),
        mode='744',
        user='root',
        group='root',
        sudo=True,
        http=host.data.http,

        state=state, host=host,
    )

    server.shell(
        name='Enable root.conf',
        commands=[
            'a2enconf root'
        ],
        sudo=True,

        state=state, host=host,
    )

    systemd.service(
        name='Restart apache2',
        service='apache2',
        running=True,
        restarted=apache_conf.changed,
        sudo=True,

        state=state, host=host,
    )


@deploy('Uninstall apache2')
def uninstall(state=None, host=None):
    supported_schema_versions = [
        v1beta3.HttpData,
    ]

    validate_schema_version(host.data.http, supported_schema_versions)

    if 'apache2.service' in host.fact.systemd_status:
        systemd.service(
            name='Stop apache2',
            service='apache2',
            running=False,
            sudo=True,

            state=state, host=host,
        )

    files.file(
        name='Remove custom config',
        path=str(Path('/etc') / 'apache2' / 'conf-available' / 'root.conf'),
        present=False,
        sudo=True,

        state=state, host=host,
    )

    apt.packages(
        name='Ensure apache2 package is not present',
        packages=['apache2'],
        present=False,
        sudo=True,

        state=state, host=host,
    )



from pathlib import Path

from pyinfra.api import deploy
from pyinfra.operations import (
    files,
    server,
)

from aircraft.deploys.ubuntu.models import v1beta3
from aircraft.deploys.ubuntu.models.v1beta3.pxe_data import InstallerType
from aircraft.validators import validate_schema_version

# This is bad. Each deploy dir should be independent. I'm just
# too lazy to copy the templates and files right now because it
# would mean modifying two copies at the same time. Some fix is
# needed down the road.
deploy_dir = Path(__file__).parent.parent / 'ubuntu'


class UnsupportedInstallerTypeError(Exception):

    def __init__(self, installer):
        msg = f"Unsupported installer type '{installer.type}'"
        super().__init__(msg)


@deploy('Configure Synology PXE')
def configure(state=None, host=None):
    supported_schema_versions = [
        v1beta3.PxeData,
    ]

    validate_schema_version(host.data.pxe, supported_schema_versions)

    #
    # Download bootfile(s)
    #

    for bootfile in host.data.pxe.bootfiles:
        bootfile_dir = (host.data.pxe.tftp.root_dir / bootfile.get_path()).parent

        files.directory(
            name='Ensure bootfile directory',
            path=str(bootfile_dir),
            present=True,
            sudo=True,

            host=host, state=state,
        )

        files.download(
            name=f'Download bootfile {bootfile.get_path()}',
            src=str(bootfile.image_source_url),
            dest=str(host.data.pxe.tftp.root_dir / bootfile.get_path()),
            sha256sum=bootfile.image_sha256sum,
            sudo=True,

            host=host, state=state,
        )

    if host.data.pxe.installer.type == InstallerType.autoinstall_v1:
        configure_installer_type_autoinstall_v1(state=state, host=host)
    elif host.data.pxe.installer.type == InstallerType.legacy_netboot:
        configure_installer_type_legacy_netboot(state=state, host=host)
    else:
        raise UnsupportedInstallerTypeError(host.data.pxe.installer)


def configure_installer_type_autoinstall_v1(state=None, host=None):

    #
    # Download the OS installer image
    #

    iso_path = host.data.pxe.http.root_dir.joinpath(
        host.data.pxe.installer.image_source_url.path.lstrip('/')
    )

    files.directory(
        name=f"Ensure {iso_path.parent}",
        path=str(iso_path.parent),
        present=True,
        sudo=True,

        host=host, state=state,
    )

    download_installer = files.download(
        name=f'Download Installer Image to {iso_path}',
        src=str(host.data.pxe.installer.image_source_url),
        dest=str(iso_path),
        sha256sum=host.data.pxe.installer.image_sha256sum,
        sudo=True,

        host=host, state=state,
    )

    # This deploy only supports serving one OS version for now and
    # to ensure that the extracted bootstrap kernel and ramdisk come
    # from the correct ISO, we use this template as one of the signals
    # in the extraction logic further down. Without this, the OS version
    # being served might change but the bootstrap kernel and ramdisk
    # might not.

    # files.template uses SFTP to transfer files so we have to use
    # a different base path in the case of Synology which presents a
    # different filesystem hierarchy depending on which protocol you're on.
    # Related bug: https://github.com/Fizzadar/pyinfra/issues/499
    current_installer_flag = host.data.pxe.tftp.sftp_root_dir / 'current-installer'
    current_installer = files.template(
        name=f'Write {current_installer_flag}',
        src=deploy_dir / 'templates' / 'current-installer.j2',
        dest=current_installer_flag,

        pxe=host.data.pxe,

        host=host, state=state,
    )

    #
    # Extract the kernel and ram disk image for use by the bootloader
    #

    kernel_path = str(host.data.pxe.tftp.root_dir / 'vmlinuz')
    initrd_path = str(host.data.pxe.tftp.root_dir / 'initrd')

    if host.fact.file(kernel_path) is None or \
       host.fact.file(initrd_path) is None or \
       download_installer.changed or \
       current_installer.changed:
        server.shell(
            name='Mount the ISO to /mnt',
            commands=[
                f'mount | grep "{iso_path} on /mnt" || mount {iso_path} /mnt',
            ],
            sudo=True,

            host=host, state=state
        )

        server.shell(
            name="Extract kernel and initrd from ISO",
            commands=[
                f'cp /mnt/casper/vmlinuz {kernel_path}',
                f'cp /mnt/casper/initrd {initrd_path}',
            ],
            sudo=True,

            host=host, state=state,
        )

    server.shell(
        name=f'Ensure {iso_path} is unmounted',
        commands=[
            f'(mount | grep "{iso_path} on /mnt" && umount /mnt) || :',
        ],
        sudo=True,

        host=host, state=state
    )

    #
    # Render GRUB2 config
    #

    # Synology's SFTP permissions are unusual in that they don't allow
    # you to create directories (which we want to do in the files.template
    # operation after this one). As a workaround to that, we're going to
    # ensure the directory via the files.directory operation since it uses
    # just SSH.
    files.directory(
        name='Ensure grub/ directory exists',
        path=str(host.data.pxe.tftp.root_dir / 'grub'),
        present=True,
        sudo=True,

        host=host, state=state,
    )

    files.template(
        name='Render GRUB config',
        src=str(deploy_dir / 'templates' / 'grub2.autoinstall-v1.cfg.j2'),
        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        dest=str(host.data.pxe.tftp.sftp_root_dir / 'grub' / 'grub.cfg'),
        create_remote_dir=False,

        pxe=host.data.pxe,
        os_name=Path(host.data.pxe.installer.image_source_url.path).stem,
        kernel_filename=Path(kernel_path).name,
        initrd_filename=Path(initrd_path).name,

        host=host, state=state,
    )

    #
    # Render the machine-specific user-data and meta-data files
    #

    for machine in host.data.pxe.machines:
        # Synology's SFTP permissions are unusual in that they don't allow
        # you to create directories (which we want to do in the files.template
        # operation after this one). As a workaround to that, we're going to
        # ensure the directory via the files.directory operation since it uses
        # just SSH.
        machine_dir = host.data.pxe.http.root_dir / machine.hostname
        files.directory(
            name=f'Ensure {machine_dir} exists',
            path=machine_dir,
            present=True,

            host=host, state=state,
        )

        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        meta_data_path = \
            host.data.pxe.http.sftp_root_dir / machine.hostname / 'meta-data'
        files.template(
            name=f'Render {meta_data_path}',
            src=str(deploy_dir / 'templates' / 'meta-data.j2'),
            dest=str(meta_data_path),
            create_remote_dir=False,
            machine=machine,

            host=host, state=state,
        )

        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        user_data_path = \
            host.data.pxe.http.sftp_root_dir / machine.hostname / 'user-data'
        files.template(
            name=f'Render {user_data_path}',
            src=str(deploy_dir / 'templates' / 'user-data.j2'),
            dest=str(user_data_path),
            create_remote_dir=False,
            machine=machine,

            host=host, state=state,
        )


def configure_installer_type_legacy_netboot(state=None, host=None):
    #
    # Download the netboot archive
    #

    archive_path = host.data.pxe.tftp.root_dir.joinpath(
        host.data.pxe.installer.netboot_source_url.path.lstrip('/')
    )

    files.directory(
        name=f"Ensure {archive_path.parent}",
        path=str(archive_path.parent),
        present=True,
        sudo=True,

        host=host, state=state,
    )

    download_netboot_archive = files.download(
        name=f'Download netboot archive to {archive_path}',
        src=str(host.data.pxe.installer.netboot_source_url),
        dest=str(archive_path),
        sha256sum=host.data.pxe.installer.netboot_sha256sum,
        sudo=True,

        host=host, state=state,
    )

    #
    # Download the OS iso
    #

    iso_path = host.data.pxe.http.root_dir.joinpath(
        host.data.pxe.installer.image_source_url.path.lstrip('/')
    )

    files.directory(
        name=f"Ensure {iso_path.parent}",
        path=str(iso_path.parent),
        present=True,
        sudo=True,

        host=host, state=state,
    )

    server.shell(
        name=f'Ensure nothing is mounted on {iso_path.parent}/mnt',
        commands=[
            f'(mount | grep " on {iso_path.parent}/mnt" && '
            f'umount {iso_path.parent}/mnt) || :',
        ],
        sudo=True,

        host=host, state=state
    )

    download_iso = files.download(
        name=f'Download OS iso to {iso_path}',
        src=str(host.data.pxe.installer.image_source_url),
        dest=str(iso_path),
        sudo=True,

        sha256sum=host.data.pxe.installer.image_sha256sum,

        host=host, state=state,
    )

    iso_mount_path = iso_path.parent / 'mnt'

    server.shell(
        name=f'Mount the ISO to {iso_path.parent}/mnt',
        commands=[
            f'mkdir -p {iso_path.parent}/mnt',
            f'mount {iso_path} {iso_path.parent}/mnt',
        ],
        sudo=True,

        host=host, state=state
    )

    # This deploy only supports serving one OS version for now and
    # to ensure that the extracted bootstrap kernel and ramdisk come
    # from the correct ISO, we use this template as one of the signals
    # in the extraction logic further down. Without this, the OS version
    # being served might change but the bootstrap kernel and ramdisk
    # might not.

    # files.template uses SFTP to transfer files so we have to use
    # a different base path in the case of Synology which presents a
    # different filesystem hierarchy depending on which protocol you're on.
    # Related bug: https://github.com/Fizzadar/pyinfra/issues/499
    current_installer_flag = host.data.pxe.tftp.sftp_root_dir / 'current-installer'
    current_installer = files.template(
        name=f'Write {current_installer_flag}',
        src=deploy_dir / 'templates' / 'current-installer.j2',
        dest=current_installer_flag,

        pxe=host.data.pxe,

        host=host, state=state,
    )

    #
    # Extract the kernel and ram disk image for use by the bootloader
    #

    kernel_path = host.data.pxe.tftp.root_dir / 'linux'
    initrd_path = host.data.pxe.tftp.root_dir / 'initrd.gz'

    if host.fact.file(kernel_path) is None or \
       host.fact.file(initrd_path) is None or \
       download_netboot_archive.changed or \
       download_iso.changed or \
       current_installer.changed:

        # We make use of the kernel and initrd in the netboot archive
        # since the the ones in the 18.04 iso, specifically under the
        # {iso_mount_path}/install directory, fail to work properly.
        kernel_path_in_archive = './ubuntu-installer/amd64/linux'
        initrd_path_in_archive = './ubuntu-installer/amd64/initrd.gz'

        server.shell(
            name='Extract the files from the netboot installer',
            commands=[
                f'tar -zxvf {archive_path} -C {kernel_path.parent} '
                f'--strip-components={kernel_path_in_archive.count("/")} '
                f'{kernel_path_in_archive}',

                f'tar -zxvf {archive_path} -C {initrd_path.parent} '
                f'--strip-components={initrd_path_in_archive.count("/")} '
                f'{initrd_path_in_archive}',
            ],
            sudo=True,

            host=host, state=state
        )

    #
    # Render Legacy Preseed Config
    #

    # files.template uses SFTP to transfer files so we have to use
    # a different base path in the case of Synology which presents a
    # different filesystem hierarchy depending on which protocol you're on.
    legacy_preseed_manual_path = host.data.pxe.http.sftp_root_dir.joinpath(
        'legacy-preseed-manual.seed'
    )
    # files.template uses SFTP to transfer files so we have to use
    # a different base path in the case of Synology which presents a
    # different filesystem hierarchy depending on which protocol you're on.
    legacy_preseed_auto_dir = host.data.pxe.http.sftp_root_dir.joinpath(
        'legacy-preseed-auto'
    )
    net_image_disk_path = iso_mount_path / 'install' / 'filesystem.squashfs'
    net_image_http_path = str(
        Path(
            host.data.pxe.installer.image_source_url.path.lstrip('/')
        ).parent.joinpath('mnt', 'install', 'filesystem.squashfs')
    )

    server.shell(
        name='Check that squashfs file exists',
        commands=[
            f'test -f {net_image_disk_path}',
        ],
        sudo=True,

        host=host, state=state
    )

    files.template(
        name='Render legacy preseed config',
        src=deploy_dir / 'templates' / 'legacy-preseed-manual.seed.j2',
        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        dest=host.data.pxe.http.sftp_root_dir / legacy_preseed_manual_path,
        create_remote_dir=False,

        pxe=host.data.pxe,
        net_image_http_path=net_image_http_path,

        host=host, state=state,
    )

    #
    # Render GRUB2 config
    #

    installer_path = Path(
        host.data.pxe.installer.netboot_source_url.path.lstrip('/')
    )

    # Synology's SFTP permissions are unusual in that they don't allow
    # you to create directories (which we want to do in the files.template
    # operation after this one). As a workaround to that, we're going to
    # ensure the directory via the files.directory operation since it uses
    # just SSH.
    grub_dir = host.data.pxe.tftp.root_dir / 'grub'
    files.directory(
        name=f'Ensure {grub_dir} exists',
        path=grub_dir,
        present=True,

        host=host, state=state,
    )

    files.template(
        name='Render GRUB config',
        src=deploy_dir / 'templates' / 'grub2.legacy-netboot.cfg.j2',
        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        dest=host.data.pxe.tftp.sftp_root_dir / 'grub' / 'grub.cfg',
        create_remote_dir=False,

        initrd_filename=Path(initrd_path).name,
        installer_path=installer_path,
        kernel_filename=kernel_path.name,
        legacy_preseed_auto_dir=legacy_preseed_auto_dir.stem,
        legacy_preseed_manual_path=legacy_preseed_manual_path.name,
        net_image_http_path=net_image_http_path,
        os_name=Path(host.data.pxe.installer.image_source_url.path).stem,
        pxe=host.data.pxe,

        host=host, state=state,
    )

    #
    # Render the machine-specific preseed files
    #

    # Synology's SFTP permissions are unusual in that they don't allow
    # you to create directories (which we want to do in the files.template
    # operation after this one). As a workaround to that, we're going to
    # ensure the directory via the files.directory operation since it uses
    # just SSH.
    legacy_preseed_auto_dir_ssh_path = \
        host.data.pxe.http.root_dir / legacy_preseed_auto_dir.stem
    files.directory(
        name=f'Ensure {grub_dir} exists',
        path=legacy_preseed_auto_dir_ssh_path,
        present=True,

        host=host, state=state,
    )

    for machine in host.data.pxe.machines:
        machine_legacy_preseed_path = legacy_preseed_auto_dir / machine.hostname
        files.template(
            name=f'Render {machine_legacy_preseed_path}',
            src=deploy_dir / 'templates' / 'legacy-preseed-auto.seed.j2',
            dest=machine_legacy_preseed_path,
            create_remote_dir=False,

            machine=machine,
            pxe=host.data.pxe,

            host=host, state=state,
        )



from enum import Enum
from pydantic import (
    AnyHttpUrl,
)
from typing import (
    List,
    Union,
)

from .base_model import BaseModel
from .bootfile_data import BootfileData
from .machine_data import MachineData
from .tftp_data import TftpData
from .http_data import HttpData


class InstallerType(str, Enum):
    autoinstall_v1 = 'autoinstall-v1'
    legacy_netboot = 'legacy-netboot'


class AutoinstallV1InstallerConfigData(BaseModel):
    type: InstallerType
    image_source_url: AnyHttpUrl
    image_sha256sum: str


class LegacyNetbootInstallerConfigData(BaseModel):
    type: InstallerType
    netboot_source_url: AnyHttpUrl
    netboot_sha256sum: str
    image_source_url: AnyHttpUrl
    image_sha256sum: str


class PxeData(BaseModel):
    tftp: TftpData
    http: HttpData

    bootfiles: List[BootfileData]
    installer: Union[
        AutoinstallV1InstallerConfigData,
        LegacyNetbootInstallerConfigData,
    ]

    machines: List[MachineData] = []

    preserve_files_on_uninstall: bool = False



from pathlib import Path
from ipaddress import (
    IPv4Address,
)
from pydantic import (
    validator,
)
from typing import (
    Optional,
)

from .base_model import BaseModel


class HttpData(BaseModel):
    """
    Holds information about where to reach the HTTP server as well as
    the path in the server where the root directory is located. The
    optional sftp_root_dir is available for systems where the SFTP
    subsystem of the SSH service presents a different path. Example
    systems that do this is Synology's DSM. If sftp_root_dir is not
    provided, then it is assumed to have the same value as root_dir.
    """
    hostname: IPv4Address
    port: Optional[int] = 80
    root_dir: Path
    sftp_root_dir: Optional[Path]

    @validator('sftp_root_dir', pre=True, always=True)
    def ensure_sftp_root_dir_has_a_value(cls, value, values):
        return value or values.get('root_dir', None)

    def get_address(self):
        """
        Returns the socket address (hostname:port) of the HTTP server
        """
        return f'{self.hostname}:{self.port}'



from pathlib import Path

from pyinfra.api import deploy
from pyinfra.operations import (
    files,
    server,
)

from aircraft.deploys.edge_os.models import v1beta3
from aircraft.validators import validate_schema_version

deploy_dir = Path(__file__).parent


@deploy('Configure EdgeOS DHCP')
def configure(state=None, host=None):
    supported_schema_versions = [
        v1beta3.DhcpData,
    ]

    validate_schema_version(host.data.dhcp, supported_schema_versions)

    filename = "dhcp-configure.sh.j2"
    file_path = './dhcp-configure.sh'

    files.template(
        name='Render configuration script',
        src=str(deploy_dir / 'templates' / filename),
        dest=file_path,
        mode='700',
        state=state, host=host,
    )

    server.shell(
        name="Execute configuration script",
        commands=[
            file_path
        ],
        state=state, host=host,
    )


@deploy('Delete EdgeOS DHCP Service')
def delete(state=None, host=None):
    supported_schema_versions = [
        v1beta3.DhcpData,
    ]

    validate_schema_version(host.data.dhcp, supported_schema_versions)

    filename = "dhcp-disable.sh.j2"
    file_path = './dhcp-disable.sh'

    files.template(
        name='Render configuration script',
        src=deploy_dir / 'templates' / filename,
        dest=file_path,
        mode='700',
        state=state, host=host,
    )

    server.shell(
        name="Execute configuration script",
        commands=[
            file_path
        ],
        state=state, host=host,
    )



from ipaddress import (
    IPv4Address,
    IPv4Interface,
)
from pydantic import (
    Field,
    validator,
)
import textwrap
from typing import (
    List,
    Union,
)
import yaml


from .base_model import BaseModel


class AddressesList(list):
    """
    Custom field for a list of IP addresses. Use this to automatically
    validate if your list contains only valid IP addresses. Example:

        from pydantic import BaseModel

        class MyNetworkConfiguration(BaseModel):
            addresses: AddressesList

        # This should raise a ValueError on the second item in the provided list
        MyNetworkConfiguration({addresses: ['1.2.4.4', 'abcd']})
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.ensure_addresses_are_valid_ips

    @classmethod
    def ensure_addresses_are_valid_ips(cls, addresses):
        for addr in addresses:
            try:
                IPv4Address(addr)
            except ValueError as e:
                raise ValueError(f"{addr} is not a valid IP address") from e
        return [str(address) for address in addresses]


class CidrAddressesList(list):
    """
    Custom field for a list of IP addresses. Use this to automatically
    validate if your list contains only valid IP addresses in CIDR notation.
    Example:

        from pydantic import BaseModel

        class MyNetworkConfiguration(BaseModel):
            addresses: CidrAddressesList

        # This should raise a ValueError on the second item in the provided list
        MyNetworkConfiguration({addresses: ['1.2.4.4/24', '1.2.4.4']})
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.ensure_addresses_are_in_cidr_notation

    @classmethod
    def ensure_addresses_are_in_cidr_notation(cls, addresses):
        for addr in addresses:
            try:
                IPv4Interface(addr)
            except ValueError as e:
                raise ValueError(f"{addr} must be in CIDR notation") from e
        return [str(address) for address in addresses]


class NetworkNameServersBaseModel(BaseModel):
    addresses: AddressesList


class NetworkDeviceBaseModel(BaseModel):
    addresses: CidrAddressesList = None
    dhcp4: bool
    gateway4: Union[str, IPv4Address] = None
    nameservers: NetworkNameServersBaseModel = None
    mtu: int = None

    @validator('gateway4')
    def ensure_gateway_is_ip_address(cls, value):
        value = str(value)
        try:
            IPv4Address(value)
        except ValueError as e:
            raise ValueError(f"{value} is not a valid IP address") from e
        return value


class NetworkBondConfigParametersData(BaseModel):
    mode: str
    lacp_rate: str = Field(alias='lacp-rate')
    mii_monitor_interval: int = Field(alias='mii-monitor-interval')


class NetworkVlanConfigData(NetworkDeviceBaseModel):
    id: int

    def export_netplan_v2(self):
        return self.dict(exclude_none=True)


class NetworkBondConfigData(NetworkDeviceBaseModel):
    name: str
    interfaces: List[str]
    parameters: NetworkBondConfigParametersData
    vlans: List[NetworkVlanConfigData] = []

    def export_netplan_v2(self):
        return self.dict(by_alias=True, exclude={'name', 'vlans'}, exclude_none=True)


class NetworkBridgeParametersData(BaseModel):
    forward_delay: int = Field(alias='forward-delay')
    priority: int
    stp: bool


class NetworkBridgeConfigData(NetworkDeviceBaseModel):
    name: str
    interfaces: List[str]
    parameters: NetworkBridgeParametersData
    vlans: List[NetworkVlanConfigData] = []

    def export_netplan_v2(self):
        return self.dict(by_alias=True, exclude={'name', 'vlans'}, exclude_none=True)


class NetworkEthernetConfigData(NetworkDeviceBaseModel):
    name: str
    vlans: List[NetworkVlanConfigData] = []

    def export_netplan_v2(self):
        return self.dict(by_alias=True, exclude={'name', 'vlans'}, exclude_none=True)


class NetworkConfigData(BaseModel):
    """
    Used for declaratively defining the network configuration of a machine.
    This model allows you to, for example, directly declare a VLAN within a
    network device such as a bond or an ethernet interface, reducing duplicate
    data and reducing errors. Example:

        network_config = NetworkConfigData({
            ethernets=[
                {
                    'name': 'eno1',
                    'dhcp4': False,
                },
                {
                    'name': 'eno2',
                    'dhcp4': False,
                }
            ],
            bonds=[
                {
                    'name': 'bond0',
                    'dhcp4': False,
                    'interfaces': [
                        'eno1',
                        'eno2'
                    ]
                    'vlans': [
                        {
                            'id': 147,
                            'dhcp4': False,
                            'addresses': [
                                '192.168.93.132/24'
                            ]
                        }
                    ],
                    'parameters': {
                        'mode': 'active-backup',
                        'primary': 'enp3s0',
                    }
                }
            ],
            bridges=[
                {
                    'name': 'br0',
                    'addresses': [
                        '192.168.94.100/24',
                    ],
                    interfaces=[
                        'bond0',
                    ]
                }
            ]
        })

        # Optionally output to a netplan v2 config:
        netplan = network_config.export_netplan_v2()
    """
    ethernets: List[NetworkEthernetConfigData] = []
    bonds: List[NetworkBondConfigData] = []
    bridges: List[NetworkBridgeConfigData] = []

    def export_netplan_v2(self, format='yaml', indent=0):
        # TODO: See if there's a cleaner way to export vlans below
        netplan_dict = {
            'network': {
                'version': 2,
                'renderer': 'networkd',
                'ethernets': {
                    ethernet.name: ethernet.export_netplan_v2()
                    for ethernet in self.ethernets
                },
                'bonds': {
                    bond.name: bond.export_netplan_v2()
                    for bond in self.bonds
                },
                'bridges': {
                    bridge.name: bridge.export_netplan_v2()
                    for bridge in self.bridges
                },
                'vlans': {
                    f"{device.name}.{vlan.id}": {
                        **vlan.export_netplan_v2(),
                        'link': device.name
                    }
                    for device in self.ethernets + self.bonds + self.bridges
                    for vlan in device.vlans
                }
            }
        }

        return textwrap.indent(yaml.dump(netplan_dict), " " * indent)



from .bootfile_data import BootfileData
from .dhcp_data import (
    DhcpData,
    DhcpRangeData,
)
from .dnsmasq_data import DnsmasqData
from .http_data import HttpData
from .machine_data import MachineData
from .network_config_data import NetworkConfigData
from .pxe_data import PxeData
from .storage_config_data import StorageConfigData
from .tftp_data import TftpData



from pydantic import (
    BaseModel as PydanticBaseModel,
)


class BaseModel(PydanticBaseModel):
    class Config:
        # Quasi-immutable model
        allow_mutation = False
        # Prevent arbitrary fields from being provided upon initialization
        extra = 'forbid'
